# ----------------------------------------------------------------
# ------------------------------------------------------------------------------
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
#  FileManager Module:
#    Classes:
#      dataTransferInterface: threaded, interface to pyGlobus.ftpClient
#      simpleServer: threaded, interface to pyGlobus.GSISOAP
#
#  status
#    NORMAL
#        0 - submitted
#        1 - committed
#        2 - active
#        3 - failed
#        4 - DONE
#    CONDITIONAL
#        5 - resource stop, partial trans
#    INTERVENTION
#        6 - admin stop
#        7 - admin stop, partial trans
#        8 - user stop
#        9 - user stop, partial trans
#
#  priority
#        0 - lo
#        1 - hi
#        2 - push
#
# ----------------------------------------------------------------
import exceptions, traceback
import sys, os, time, socket, getopt, threading, ConfigParser, urlparse

from NetworkService import *
from DataManager import *
from Util import *

import pyGlobus
from pyGlobus import ftpClient
from pyGlobus.ftpControl import Parallelism, TcpBuffer
from pyGlobus.GSISOAP import SOAPServer, MethodSig, faultType, ioc
from pyGlobus.security import GSSCred, import_external
from pyGlobus.util  import GlobusException
from exceptions import Exception

#default codes
CRED_ERR   = ('401', 'credential error') 
INPUT_ERR  = ('402', 'input error') #ValueError
STATUS_ERR = ('403', 'status invalidates operation')
DATA_ERR   = ('404', 'possible data corruption') #RuntimeError

#globals
transM = None
debug = None

# ----------------------------------------------------------------
#  TransferManager:
#      Initiale all DTI threads.
#      _dTI_list: List of all dTI's
#      _handle_dTI-dir: Directory of active dTI's
# ----------------------------------------------------------------
class TransferManager:
    def __init__(self, concurrent_jobs, tcp_streams, tcp_buffer, nwsclass=None, **plugins):
        self._handle_dTI_dir = {}
        self._dTI_list = []
        self._nwsclass = nwsclass
        self._plugins = plugins
        for i in range(concurrent_jobs):
            dTrans = DataTransferInterface(tcp_streams, tcp_buffer, str(i), )
            self._dTI_list.append(dTrans)
            dTrans.start()
        Log_threads(5)
        return
    # ----------------------------------------------------------------
    #  start:  Once we've taken a job off the 'queue' it is
    #     registered as an active job, this info is passed
    #     to an underlying thread
    #
    # ----------------------------------------------------------------
    def start(self):
        dataInt = get_DataInterface()
        while len(self._dTI_list):
            trans = dataInt.get_next_job()
            while trans:
                print "** Start Transfer"
                for dTrans in self._dTI_list:
                    old_handle = dTrans.get_handle()
                    try:
                        isReady = dTrans.init_transfer(self._nwsclass, *trans, **self._plugins)
                    except ValueError, ex:
                        Log_exception(ex)
                        Log_error(('TransferManager received bad job'))
                        dataInt.transfer_failed(trans[0])
                    else:
                        if isReady:
                             if old_handle:
                                 Log_debug(("old_handle = %s" %(old_handle)),4)
                                 del self._handle_dTI_dir[old_handle]
                             new_handle = dTrans.get_handle()
                             self._handle_dTI_dir.update({new_handle:dTrans})
                             trans = None
                             break
                else:
                    Log_debug(('TransferManager job waiting for a thread'), 4)
                    time.sleep(0.5)
            else:
                time.sleep(0.5)
        else:
            Log_debug(('TransferManager transfers will not occur'), 4)
        return
    def stop(self, handle):
        dTI = self._handle_dTI_dir.get(handle)
        if dTI:
            try:
                dTI.stop_transfer()
            except KeyError, ex:
                raise ValueError(('stop transfer fail, no ftpClient for handle'))
            else:
                dTI.stop_transfer(handle)
                del self._handle_dTI_dir.update[handle]
                Log_debug(('stop transfer (%s)' %(handle)))
        return

# ----------------------------------------------------------------
#  DataTransferInterface
#
#   dname =
#     '/O=Grid/O=Lawrence Berkeley National Laboratory/CN=host/portnoyc.lbl.gov'
#   sname =
#     '/O=Grid/O=Lawrence Berkeley National Laboratory/CN=host/scott.lbl.gov'
#   user  = 'boverhof'
#   srcAttr.set_authorization(newCred, user, '', '', sname)
#   dstAttr.set_authorization(newCred, user, '', '', dname)
#
# ---------------------------------------------------------------
class DataTransferInterface(threading.Thread):
    def __init__(self, tcp_streams, tcp_buffer, thread_name=None):
        threading.Thread.__init__(self, None, thread_name)
        self._tcp_streams = tcp_streams
        self._tcp_buffer_size = tcp_buffer
        self._ftpClient = None
        self._handle = None
        self._src_url = None
        self._dst_url = None
        self._marker = None
        self._credPath = None
        self._condV =  threading.Condition(threading.Lock())
        self._active = 0
        self._plugins = None
        return
    def to_string(self):
        dti = (('tcp_streams = %d, tcp_buffer_size = %d, handle = %s,\n' +
                'src = %s, dst = %s, marker = %s, cred = %s, active = %d\n')
               %(self._tcp_streams, self._tcp_buffer_size, self._handle,
                 self._src_url, self._dst_url, self._marker, self._credPath, 
                 self._active))
        return dti
    def get_handle(self):
        return self._handle
    def set_buffer_size(self, tcp_buffer):
        self._tcp_buffer_size = tcp_buffer
        return
    def set_tcp_streams(self, tcp_streams):
        self._tcp_streams = tcp_streams
        return
    # ---------------------------------------------------------------
    # init_transfer(self, trans_tuple):
    #    success: return old_handle
    #    failure: return 0
    # ---------------------------------------------------------------
    def init_transfer(self, nwsclass, *trans_tuple, **plugins):
        started = 0
        if not self._active:
             self._condV.acquire(1)
             self._active = 1
             self._plugins = plugins
             started = 1 
             print trans_tuple
             self._handle, self._src_url, self._dst_url,\
                           self._marker, self._credPath = trans_tuple
             if nwsclass:
                 up = urlparser()
                 src = up.urlparse(self._src_url)
                 dst = up.urlparse(self._dst_url)
                 for  serv, dest in ((src[1], dst[1]), (dst[1], src[1])):
                     try:
                         ns = (eval(nwsclass))(serv)
                         buf = ns.getBufferSize(dest)
                     except RuntimeError, ex:
                         Log_exception(ex)
                     except Exception, ex:
                         Log_exception(ex)
                     else:
                         Log_debug(("buffersize = %d" %(buf)), 3)
                         if buf > 0:
                             self._tcp_buffer_size = buf
                             break
                 else:
                     Log_event(('Network Service Failed'))
             self._condV.notify()
             self._condV.release()
        return started
    # ---------------------------------------------------------------
    # run: catch all exceptions
    # ---------------------------------------------------------------
    def run(self):
        condV =  threading.Condition(threading.Lock())
        while 1:
            Log_debug(("thread ready: %s" %(self.getName())), 5)
            self._condV.acquire(1)
            self._active = 0
            self._condV.wait()
            Log_debug(("*** %s -transfer-> %s ***"
                       %(self.getName(), self._handle)), 4)
            Log_threads(5)
            dataInt = get_DataInterface()
            Log_debug(self.to_string(), 5)
            try:
                newCred = pyGlobus.security.import_external(self._credPath)
            except pyGlobus.security.GSSContextException, ex:
                dataInt.transfer_failed(self._handle)
                Log_exception(ex)
            else:
                #-------------- Set Handle Attributes -----------------
                handleAttr = ftpClient.HandleAttr()
                print self._plugins
                if self._plugins.get(('RestartMarkerPlugin')):
                    print "RestartMarkerPlugin READY"
                    try:
                        rm_plugin = ftpClient.RestartMarkerPlugin(beginCB,
                                                                  markerCB,
                                                                  completeCB,
                                                                  self._handle)
                    except ftpClient.PluginException, ex:
                        Log_exception(ex)
                    else:
                        handleAttr.add_plugin(rm_plugin)
                if self._plugins.get(('DebugPlugin')):  
                    try:
                        debug_plugin = ftpClient.DebugPlugin(sys.stdout,
                                                             ('DebugPlugin'))
                    except ftpClient.PluginException, ex:
                        Log_exception(ex)
                    else:
                        handleAttr.add_plugin(debug_plugin)
                if self._plugins.get(('ThroughputPerfPlugin')):
                    print "ThoughputPerfPlugin READY"
                    try:
                        tp_plugin = ftpClient.ThoughputPerfPlugin(beginTP_CB,
                                                                  stripeTP_CB,
                                                                  totalTP_CB,
                                                                  completeTP_CB,
                                                                  self._handle)
                    except ftpClient.PluginException, ex:
                        Log_exception(ex)
                    else:
                        handleAttr.add_plugin(tp_plugin)
                elif self._plugins.get(('PerformanceMarkerPlugin')):
                    print "PerformanceMarkerPlugin READY"
                    try:
                        pm_plugin = ftpClient.PerformanceMarkerPlugin(beginPM_CB,
                                                                      markerPM_CB,
                                                                      completePM_CB,
                                                                      self._handle)
                    except ftpClient.PluginException, ex:
                        Log_exception(ex)
                    else:
                        handleAttr.add_plugin(pm_plugin)
                self._ftpClient = ftpClient.FtpClient(handleAttr)
                #-------------- Set Operation Attributes --------------
                try:
                    parallelism = Parallelism()
                    parallelism.set_mode(1)
                    parallelism.set_size(self._tcp_streams)
                except ParallelismException, ex:
                    Log_exception(ex)
                    parallelism = None
                try:
                    tcp_buffer = TcpBuffer()
                except TcpBufferException, ex:
                    Log_exception(ex)
                    tcp_buffer = None
                else:
                    if self._tcp_buffer_size > 0:
                        tcp_buffer.set_fixed(self._tcp_buffer_size)
                    else:
                        tcp_buffer.set_default()
                        
                srcAttr = ftpClient.OperationAttr()
                dstAttr = ftpClient.OperationAttr()
                for attr in (srcAttr, dstAttr):
                    attr.set_mode('E')
                    attr.set_authorization(newCred)
                    if parallelism:
                        attr.set_parallelism(parallelism)
                    if tcp_buffer:
                        attr.set_tcp_buffer(tcp_buffer)
                #-------------- Perform Transfer ----------------------
                condV.acquire(1)
                try:
                    cb = self._ftpClient.third_party_transfer(
                        self._src_url, self._dst_url, done_func,
                        condV, srcAttr, dstAttr)
                except ftpClient.FtpClientException, ex:
                    Log_exception(ex)
                    condV.release()
                    dataInt.transfer_failed(self._handle)
                except Exception, ex:
                    Log_error("thrid_party_transfer")
                    Log_exception(ex)
                    condV.release()
                    dataInt.transfer_failed(self._handle)
                else:
                    condV.wait()
                    condV.release()
            self._condV.release()
        return
    # ----------------------------------------------------------------
    #  stop_transfer:
    #
    # ----------------------------------------------------------------
    def stop_transfer(self):
        if self._ftpClient:
             self._ftpClient.abort()
        else:
             raise ValueError, 'ftpClient is not initialized'
        return
        
# ----------------------------------------------------------------
# Callback Functions: Used by dataTransferInterface class
#
# ----------------------------------------------------------------
def done_func(cv, marker, err):
    Log_debug(("done_func\n\tmarker = %s\n\terror=%s" %(marker, err)), 4)
    cv.acquire(1)
    cv.notify()
    cv.release()
    return
# --------------------- RestartMarkerPlugin Callbacks ------------
def beginCB(handle, ftp_handle, src_url, dst_url, marker):
    dataInt = get_DataInterface()
    ret = 0
    msg = "beginCB\n\thandle = %s\n\tftp_handle = %s\n\tsrc = %s" \
          + "\n\tdest = %s\n\tmarker = %s" 
    Log_debug((msg %(handle, ftp_handle, src_url, dst_url, marker)), 4)
    try:
        Log_netlogger(("TRANSFER_START"),
                      ("handle=%(handle)s ftp_handle=%(ftp_handle)s src_url=%(src_url)s dst_url=%(dst_url)s" %vars()))
    except Exception, ex:
        print "ERROR Log_netlogger ", ex
    del msg
    try:
        marker_str = dataInt.get_marker(handle)
    except ValueError, ex:
        Log_exception(ex)
    else:
        if marker_str:
            Log_debug(("marker_str: %s" %(marker_str)), 4)
            try:
                ftpClientc.restart_marker_from_string(marker, marker_str)
            except ftpClient.FtpClientException, ex:
                Log_exception(ex)
            else:
                ret = 1
    return ret
def markerCB(handle, ftp_handle, marker):
    dataInt = get_DataInterface()
    Log_debug(("markerCB\n\thandle = %s\n\tftp_handle = %s\n\ttmarker = %s"
                 %(handle, ftp_handle, marker)), 4)
    ret, err, marker_str = ftpClientc.restart_marker_to_string(marker)
    Log_debug(("restart_marker\n\tret = %d\n\terr = %s\n\tmarker_str = %s"
                 %(ret, err, marker_str)), 4)
    dataInt.set_marker(handle, marker_str)
    return
def completeCB(handle, ftp_handle, err, url):
    dataInt = get_DataInterface()
    msg = "completeCB\n\thandle = %s\n\tftp_handle = %s" \
          + "\n\terr = %s\n\turl = %s"
    Log_debug((msg %(handle, ftp_handle, err, url)), 4)
    try:
        Log_netlogger(("TRANSFER_END"),
                      ("handle=%(handle)s ftp_handle=%(ftp_handle)s err=%(err)s url=%(url)s" %vars()))
    except Exception, ex:
        print ex
    del msg
    if err != "NONE": dataInt.transfer_failed(handle)
    else:             dataInt.transfer_done(handle)
    return
# --------------------- PerformanceMarkerPlugin Callbacks --------
def beginPM_CB(handle, ftp_handle, src_url, dst_url, restart):
    dataInt = get_DataInterface()
    msg = "beginPM_CB\n\thandle(%s) ftp_handle(%s) src(%s) dst(%s)" 
    Log_debug((msg %(handle, ftp_handle, src_url, dst_url, restart)), 3)
    del msg
    return
def markerPM_CB(handle, ftp_handle, time_stamp, time_stampL, stripe_ndx,
                num_stripes, nbytes):
    print "*** markerPM_CB *** "
    dataInt = get_DataInterface()
    msg = "handle(%s) ftp_handle(%s)\n\ttime(%d) stripe(%d) num(%d) bytes(%d)"
    Log_debug((msg %(handle, ftp_handle, time_stamp, stripe_ndx, num_stripes, nbytes)), 4)
    del msg
    try:
        Log_netlogger(("PERFORMANCE_MARKER"),
                      ("handle=%(handle)s ftp_handle=%(ftp_handle)s time_stamp=%(time_stamp)d time_stampL=%(time_stampL)c stripe_ndx=(stripe_ndx)d num_stripes=%(num_stripes)d bytes=%(nbytes)d" %vars()))
    except Exception, ex:
        print "ERROR Log_netlogger ", ex
    return
def completePM_CB(handle, ftp_handle, success):
    dataInt = get_DataInterface()
    msg = "completePM_CB\n\thandle(%s) ftp_handle(%s) success(%s)"
    Log_debug((msg %(handle, ftp_handle, success)), 3)
    del msg
    return
# --------------------- ThroughPut Plugin Callbacks --------
def beginTP_CB(handle, ftp_handle, src_url, dst_url):
    dataInt = get_DataInterface()
    msg = "beginTP_CB\n\thandle(%s) ftp_handle(%s) src(%s) dst(%s)" 
    Log_debug((msg %(handle, ftp_handle, src_url, dst_url)), 3)
    del msg
    return
def stripeTP_CB(handle, ftp_handle, stripe_ndx, bytes, inst_thrput, avg_thrput):
    print "*** stripeTP_CB *** "
    dataInt = get_DataInterface()
    msg = (("stripeTP_CB\n\thandle(%s) ftp_handle(%s)\n\t" +
            "strip_ndx(%d) bytes(%d) inst(%d bytes/sec) avg(%d bytes/sec)")
           %(handle, ftp_handle, stripe_ndx, bytes, inst_thrput, avg_thrput))
    Log_debug(msg , 4)
    try:
        Log_netlogger(("STRIPE_THROUGHPUT"),
                      ("handle=%(handle)s ftp_handle=%(ftp_handle)s stripe=(stripe_ndx)d bytes=%(bytes)d instant=%(inst_thrput)d ave=%(avg_thrput)d" %vars()))
    except Exception, ex:
        print "ERROR Log_netlogger ", ex
    del msg
    return
def totalTP_CB(handle, ftp_handle, bytes, inst_thrput, avg_thrput):
    print "*** totalTP_CB *** "
    dataInt = get_DataInterface()
    msg = (("totalTP_CB\n\thandle(%s) ftp_handle(%s)\n\t" +
            "bytes(%d) inst(%d bytes/sec) avg(%d bytes/sec)")
           %(handle, ftp_handle, bytes, inst_thrput, avg_thrput))
    Log_debug(msg , 4)
    try:
        Log_netlogger(("TOTAL_THROUGHPUT"),
                      ("handle=%(handle)s ftp_handle=%(ftp_handle)s bytes=%(bytes)d instant=%(inst_thrput)d ave=%(avg_thrput)d" %vars()))
    except Exception, ex:
        print "ERROR Log_netlogger ", ex
    del msg
    return
def completeTP_CB(handle, ftp_handle, success):
    dataInt = get_DataInterface()
    msg = "completeTP_CB\n\thandle(%s) ftp_handle(%s) success(%s)"
    Log_debug((msg %(handle, ftp_handle, success)), 3)
    del msg
    return

# ----------------------------------------------------------------
# ----------------------------------------------------------------
#  testServer:
#
#  status
#    NORMAL
#        0 - submitted
#        1 - committed
#        2 - active
#        3 - failed
#        4 - DONE
#    CONDITIONAL
#        5 - resource stop, partial trans
#    INTERVENTION
#        6 - admin stop
#        7 - admin stop, partial trans
#        8 - user stop
#        9 - user stop, partial trans
#
# ----------------------------------------------------------------
# ----------------------------------------------------------------
#  class simpleServer: register functions and start up a server
#      thread.
#
# ----------------------------------------------------------------
class SOAPServerX(SOAPServer):
    def __init__(self, server_address):
        SOAPServer.__init__(self, server_address)
    def handle_request(self):
        """Handle one request, possibly blocking."""
        try:
            request, client_address = self.get_request()
        except socket.error, ex:
            Log_exception(ex)
        except Exception, ex:
            Log_exception(ex)
        else:
            print "HI"
            if self.verify_request(request, client_address):
                try:
                    self.process_request(request, client_address)
                except Exception, ex:
                    Log_debug(("hr process_request"), 4)
                    self.handle_error(request, client_address)
            #??? 
            #self.close_request(request)
            Log_debug(("request: %s\n" %(str(request))), 4)
            Log_debug(("client:  %s\n" %(str(client_address))), 4)
        return
#    def get_request(self):
#        return self.socket.accept()
        
class simpleServer(threading.Thread):
    def __init__(self, my_name, port, namespace):
        #self._server = SOAPServer((my_name, port))
        self._server = SOAPServerX((my_name, port))
        self._server.channel_mode = ioc.GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP
        self._server.delegation_mode = (
            ioc.GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY)
        self._server.namespace = namespace
        self._mF = SoapServerMethods()
        self._server.registerFunction(MethodSig(self._mF.printDS,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.getDump,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.submitTransferRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.commitTransferRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.completeTransferRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.listTransferRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.checkStatusRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.changePriorityRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.clearRecordRequest,
                                                keywords=0, context=1))
        self._server.registerFunction(MethodSig(self._mF.userStopTransferRequest,
                                                keywords=0, context=1))
        
        threading.Thread.__init__(self)
        self.setName('SOAPServer')
        return
    def run(self):
        while 1:
            try:
                self._server.handle_request()
            except socket.error, ex:
                Log_exception(ex)
            except Exception, ex:
                Log_exception(ex)
            else:  Log_debug(("simpleServer received request"), 4)
        return

class SoapServerMethods:
    def __init__(self):
        return
    # ----------------------------------------------------------------
    #  printDS: print datastructures
    # ----------------------------------------------------------------
    def printDS(self, _SOAPContext):
        dataInt = get_DataInterface()
        Log_event(dataInt.to_string())
        return
    # ----------------------------------------------------------------
    #  getDump:
    # ----------------------------------------------------------------
    def getDump(self, _SOAPContext):
        dataInt = get_DataInterface()
        dump = dataInt.to_string()
        Log_debug(("-- getDump\n\tdump - %s" %(dump)), 4)
        return dump
    # ----------------------------------------------------------------
    #  submitTransferRequest: make request record in Database
    #    -check URL at this level
    # ----------------------------------------------------------------
    def submitTransferRequest(self, url_src, url_dst, priority, _SOAPContext):
        msg = "-- submitTransferRequest\n\tsrc - %s" \
              + "\n\tdst - %s\n\tpriority - %s"
        Log_debug((msg %(url_src, url_dst, priority)), 4)
        del msg
        h = self.completeTransferRequest(None, url_src, url_dst, priority,
                                    _SOAPContext)
        return h
    # ----------------------------------------------------------------
    #  completeTransferRequest: make request record in Database
    #    -check URL at this level
    # ----------------------------------------------------------------  
    def completeTransferRequest(self, handle, url_src, url_dst, priority,
                                _SOAPContext):
        dataInt = get_DataInterface()
        c = _SOAPContext
        msg = "-- completeTransferRequest\n\tsrc - %s\n\t" \
              + "dst - %s\n\tpriority - %s"
        Log_debug((msg %(url_src, url_dst, priority)), 4)
        del msg
        try:
            cred = c.delegated_cred
            credPath = cred.export_external()
            newCred = import_external(credPath)
            h = dataInt.submit_job(url_src, url_dst, int(priority),
                                   credPath, handle)
        except AttributeError, ex:
            Log_exception(ex)    
            raise faultType(CRED_ERR[0], ((CRED_ERR[1],) + ex.args))
        except GlobusException, ex:
            Log_exception(ex)
            raise faultType(CRED_ERR[0], ((CRED_ERR[1],) + ex.args))
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        else:
            msg = "\tcred handle - %s\n\tcredPath - %s\n\timport cred handle - %s"
            Log_debug((msg %(cred.get_handle(), credPath, newCred._handle)), 4)
            del msg
        if handle:
            h = self.commitTransferRequest(h, c)
        return h
    # ----------------------------------------------------------------
    #  commitTransferRequest: update database record, put request in
    #     appropriate Queue
    #
    # ----------------------------------------------------------------
    def commitTransferRequest(self, handle, _SOAPContext):
        dataInt = get_DataInterface()
        Log_debug(("-- commitTransferRequest\n\thandle - %s" %(handle)), 4)
        try:
            h = dataInt.commit_job(handle)
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        except RuntimeError, ex:
            Log_exception(ex)
            raise faultType(DATA_ERR[0], ((DATA_ERR[1],) + ex.args))
        except StatusError, ex:
            Log_exception(ex)
            raise faultType(STATUS_ERR[0], ((STATUS_ERR[1],) + ex.args))
        return h
    # ----------------------------------------------------------------
    #  listTransferRequest:  Return record of handle in Queue
    #
    # ----------------------------------------------------------------
    def listTransferRequest(self, handle, _SOAPContext):
        dataInt = get_DataInterface()
        Log_debug(("-- listTransferRequest\n\thandle = %s" %(handle)), 4)
        try:
            url_src, url_dst, mrk, stat, prior = dataInt.get_record(handle)
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        except RuntimeError, ex:
            Log_exception(ex)
            raise faultType(DATA_ERR[0], ((DATA_ERR[1],) + ex.args))
        else:
            trans_str = "handle:%s/status:%d/priority:%d/src:%s/dst:%s/marker:%s"\
                        %(handle, stat, prior, url_src, url_dst, mrk)
        Log_debug(trans_str, 4)
        return trans_str
    # ----------------------------------------------------------------
    #  checkStatusRequest: Returns status from database entry
    #
    # ----------------------------------------------------------------
    def checkStatusRequest(self, handle, _SOAPContext):
        dataInt = get_DataInterface()
        Log_debug(("-- checkStatusRequest\n\thandle = %s" %(handle)), 4)
        try:
            s = dataInt.get_status(handle)
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        except RuntimeError, ex:
            Log_exception(ex)
            raise faultType(DATA_ERR[0], ((DATA_ERR[1],) + ex.args))
        return str(s)
    # ----------------------------------------------------------------
    #  changePriorityRequest: only status 0 and 1 can be moved
    #     status
    #        0 - record in database
    #        1 - record in database, Queue
    #     new_priority
    #        0 - move to back of _loQueue
    #        1 - move to back of _hiQueue
    #        2 - move to top of _hiQueue (PUSH)
    #
    #    returns 'handle' or if no change ''
    #
    # ----------------------------------------------------------------
    def changePriorityRequest(self, handle, priority,_SOAPContext):
        dataInt = get_DataInterface()
        Log_debug(("-- changePriorityRequest\n\thandle = %s\n\tnew priority = %s"
                   %(handle, priority)), 4)
        try:
            h = dataInt.change_priority(handle, int(priority))
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        except StatusError, ex:
            Log_exception(ex)
            raise faultType(STATUS_ERR[0], ((STATUS_ERR[1],) + ex.args))
        except RuntimeError, ex:
            Log_exception(ex)
            raise faultType(DATA_ERR[0], ((DATA_ERR[1],) + ex.args))
        return h
    # ----------------------------------------------------------------
    #  clearRecordRequest:  clear record of an inactive job
    #    status: STOP,DONE -> X
    # ----------------------------------------------------------------
    def clearRecordRequest(self, handle,_SOAPContext):
        dataInt = get_DataInterface()
        Log_debug(("-- clearRecordRequest\n\thandle = %s " %(handle)), 4)
        try:
            h = dataInt.clear_record(handle)
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        except StatusError, ex:
            Log_exception(ex)
            raise faultType(STATUS_ERR[0], ((STATUS_ERR[1],) + ex.args))
        except RuntimeError, ex:
            Log_exception(ex)
            raise faultType(DATA_ERR[0], ((DATA_ERR[1],) + ex.args))
        return h
    # ----------------------------------------------------------------
    #  userStopTransferRequest
    #   status ACTV -> 
    # ----------------------------------------------------------------
    def userStopTransferRequest(self, handle,_SOAPContext):
        global transM
        dataInt = get_DataInterface()
        Log_debug(("-- userStopTransferRequest\n\thandle = %s " %(handle)), 4)
        try:
            h = dataInt.user_stop_active(handle)
            transM.stop(handle)
        except ValueError, ex:
            Log_exception(ex)
            raise faultType(INPUT_ERR[0], ((INPUT_ERR[1],) + ex.args))
        except StatusError, ex:
            Log_exception(ex)
            raise faultType(STATUS_ERR[0], ((STATUS_ERR[1],) + ex.args))
        except RuntimeError, ex:
            Log_exception(ex)
            raise faultType(DATA_ERR[0], ((DATA_ERR[1],) + ex.args))
        return h
    # ----------------------------------------------------------------
    #  printContextInfo: context object has extra info about the call
    #   c.xmldata         - The original XML request
    #   c.header          - The SOAP Header or None if not present
    #   c.header.mystring - An element of the SOAP Header
    #   c.body            - The whole Body object
    #   c.connection.getpeername() - The socket object, useful for
    #   c.soapaction      - The SOAPaction HTTP header
    #   c.httpheaders     - All the HTTP headers
    # ----------------------------------------------------------------
    def printContextInfo(self, _SOAPContext):
        c = _SOAPContext
        sep = '-' * 72
        x = ("-- XML%s\n%s\n-- Header%s\n%s\n-- Body%s\n%s\n"
             + "-- Peer%s\n%s\n-- SOAPAction%s\n%s\n-- HTTP headers%s\n%s\n") \
             %(sep[7:], c.xmldata, sep[10:], c.header, sep[8:], c.body, sep[8:],
               c.connection.getpeername(), sep[14:], c.soapaction, sep[16:],
               c.httpheaders)
        if c.header.mystring:
            x += (("-- Header.mystring%s\n%s\n" %(sep[19:], c.header.mystring)))
        Log_event(x)
        return

# ----------------------------------------------------------------
# ----------------------------------------------------------------
#  main
# ----------------------------------------------------------------
def main():
    global debug
    #----------------- set defaults -----------------
    cp = ConfigParser.ConfigParser()
    cp.readfp(open('defaults.cfg'))
    cp_section = 'Server'
    transfer_enable = cp.get(cp_section, 'transfer_enable')
    database        = cp.get(cp_section, 'database')
    db_name         = cp.get(cp_section, 'db_name')
    db_dir          = cp.get(cp_section, 'db_dir')
    db_table        = cp.get(cp_section, 'db_table')
    namespace       = cp.get(cp_section, 'namespace')
    log_file        = cp.get(cp_section, 'log_file')
    netlog_url      = cp.get(cp_section, 'netlog_url')
    nwsclass        = cp.get(cp_section, 'nwsclass')
    plugins         = cp.get(cp_section, 'plugins')
    plugins         = eval(plugins)
    if type(plugins) != type({}):
        raise ValueError, 'bad value in defaults.cfg, expected %s for plugins' %(type({}))
    port            = cp.getint(cp_section, 'port')
    max_jobs        = cp.getint(cp_section, 'max_jobs')
    tcp_streams     = cp.getint(cp_section, 'parallel_streams')
    tcp_buffer      = cp.getint(cp_section, 'tcp_buffer_size')
    debug           = cp.getint(cp_section, 'debug_level')
    del cp
    del cp_section
    if tcp_streams <= 0:
        raise RuntimeError(('configuration error'),
                           ('parallel_streams must be greater than zero'))
    #----------------- Initialize -------------------
    my_name = socket.gethostname()
    try:
        opt, arg = getopt.getopt(sys.argv[1:], 'd:h:p:t:n:')
    except getopt.error:
        usage(sys.argv)
    for o,a  in opt:
        if   o == '-h': my_name = a
        elif o == '-p': port = int(a)
        elif o == '-t': transfer_enable = int(a)
        elif o == '-d': debug = int(a)
        elif o == '-n': netlog_url = a
        elif o == '-e': nwsclass = a
        else: usage(sys.argv)
        
    l = Logger(log_file, netlog_url, (my_name + ':' + str(port)), debug)
    try:
        my_name = socket.gethostbyaddr(my_name)
    except NameError, ex:
        Log_exception(ex)
        sys.exit(-1)

    dI = DataInterface(database, db_name, db_dir, db_table)
    #----------------- Server Thread -----------------
    try:
        serverThread = simpleServer(my_name[0], port, namespace)
    except pyGlobus.io.GSITCPSocketException, ex:
        Log_exception(ex)
        sys.exit(-1)
    serverThread.start()
    Log_threads(5)
    #----------------- Launch job threads -----------------
    if transfer_enable:
        Log_debug(("main: start transfer"), 4)
        global transM
        transM = TransferManager(max_jobs, tcp_streams, tcp_buffer, nwsclass, **plugins)
        transM.start()
    else:
        while 1:
            time.sleep(3)
    return
# ----------------------------------------------------------------
#  usage: Called by main
# ----------------------------------------------------------------
def usage(argList):
    Log_event("usage:\n" + argList[0] + '\n')
    for o,a in argList[1:]:
        if o[0] == '-' and a[0] == '-':
            Log_event('\t' + o + ' ' + a + '\n')
        else:
            Log_event('\t' + o + '\n\t' + a + '\n')
    sys.exit(1)

if __name__=='__main__':
    main()
